﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;

namespace Sharing
{
	class TermGraph
	{
		public readonly List<Node> Nodes = new List<Node>();
		public int Count => Nodes.Count;

		public Node AddApp(Node left, Node right)
		{
			Node node = new Node(Node.NodeType.App);
			node.AddLeft(left);
			node.AddRight(right);
			Nodes.Add(node);
			return node;
		}

		public Node AddLam(Node body)
		{
			Node node = new Node(Node.NodeType.Lam);
			node.AddBody(body);
			Nodes.Add(node);
			return node;
		}

		public Node AddVar()
		{
			Node node = new Node(Node.NodeType.Var);
			Nodes.Add(node);
			return node;
		}

		public void AddUndirectEdge(Node left, Node right)
		{
			UndirectedEdge edge = new UndirectedEdge(left, right);
			left.AddUndirectedEdge(edge);
			right.AddUndirectedEdge(edge);
		}

		public TermGraph Clone()
		{
			TermGraph graph = new TermGraph();
			Dictionary<Node, Node> map = new Dictionary<Node, Node>();

			foreach (Node node in Nodes)
			{
				map[node] = new Node(node.Type);
				graph.Nodes.Add(map[node]);
			}

			foreach (Node node in Nodes)
				node.Copy(map);

			return graph;
		}

		public Node DuplicateNode(Node node)
		{
			switch (node.Type)
			{
				case Node.NodeType.App:
				{
					Node left = DuplicateNode(node.Left);
					return AddApp(left, node.Right == node.Left ? left : DuplicateNode(node.Right));
				}
				case Node.NodeType.Lam:
				{
					return AddLam(DuplicateNode(node.Body));
				}
				case Node.NodeType.Var:
				{
					if (node.Binder == null)
						return node;
					Node v = AddVar();
					v.Binder = node.Binder;
					return v;
				}
				default: throw new Exception();
			}
		}

		private int file = 0;
		private string path = @"/usr/bin/dot";
		//private string path = @"release\bin\dot";


		public void Print()
		{
			List<string> message = new List<string>() { "digraph {" };

			for (int i = 0; i < Nodes.Count(); i++)
				//if (Nodes[i].IsAlive)
				{
					Nodes[i].Index = i;
					message.Add(Nodes[i].Index + " [ ordering=out label=\"" + Nodes[i].Label + "\"" + " ]");
				}

			for (int i = 0; i < Nodes.Count(); i++)
				//if (Nodes[i].IsAlive)
				{
					if (Nodes[i].Left != null && Nodes[i].Type == Node.NodeType.Var)
						message.Add(Nodes[i].Index + " -> " + Nodes[i].Left.Index + " [constraint=false]");
					else if (Nodes[i].Left != null)
						message.Add(Nodes[i].Index + " -> " + Nodes[i].Left.Index );

					if (Nodes[i].Right != null)
						message.Add(Nodes[i].Index + " -> " + Nodes[i].Right.Index );

					foreach (UndirectedEdge edge in Nodes[i].UndirectedEdges)
						if (edge.Left == Nodes[i])
							message.Add(edge.Left.Index + " -> " + edge.Right.Index + " [dir=none constraint=false color=blue]");
					
					if (Nodes[i].Canonic != null)
						message.Add(Nodes[i].Index + " -> " + Nodes[i].Canonic. Index + " [constraint=false color=red]");
				}
				
			message.Add("}");

			File.WriteAllLines(file.ToString() + ".txt", message);
			Process.Start(path, "-O -T pdf " + file + ".txt");
			file++;

			Console.WriteLine();
		}
	}
}
