﻿using System;
using System.Collections.Generic;
using System.Linq;

namespace Sharing
{
	class TermGraphBuilder
	{
		public static void Enumerate(Action<TermGraph> action)
		{
			for (int nodes = 1; ; nodes++)
				Enumerate(nodes, action);
		}

		public static void Enumerate(int nodes, Action<TermGraph> action)
		{
			TermGraph graph = new TermGraph();
			EnumerateNodes(nodes, graph, root =>
			{
				/*if (graph.Count == nodes && !graph.Nodes.Any(app => app.Type == Node.NodeType.App &&
					(/ *app.Left.DescendantVars().Any(v => v.Binder != null && !app.DominatedBy(v.Binder)) ||* /
					app.Right.DescendantVars().Any(v => v.Binder != null && !app.DominatedBy(v.Binder)))))
					action(graph);*/
				if (graph.Count == nodes && !graph.Nodes.Any(v => v.Type == Node.NodeType.Var && v.Binder != null && !v.DominatedBy(v.Binder)))
					action(graph);
			});
		}

		private static void EnumerateNodes(int nodes, TermGraph graph, Action<Node> action)
		{
			// TODO: duplicate cases (?)
			if (graph.Count < nodes)
			{
				Node var = graph.AddVar();
				action(var);
				graph.Nodes.Remove(var);

				Node app = new Node(Node.NodeType.App);
				graph.Nodes.Add(app);
				EnumerateNodes(nodes/* - 1*/, graph, left =>
				{
					if (!app.HasAncestor(left))
						EnumerateNodes(nodes, graph, right =>
						{
							if (!app.HasAncestor(right))
							{
								app.AddLeft(left);
								app.AddRight(right);
								action(app);
								app.RemoveRight();
								app.RemoveLeft();
							}
						});
				});
				graph.Nodes.Remove(app);

				if (graph.Count < nodes - 1)
				{
					Node lam = new Node(Node.NodeType.Lam);
					graph.Nodes.Add(lam);
					EnumerateNodes(nodes, graph, right =>
					{
						if (!lam.HasAncestor(right))
						{
							lam.AddRight(right);
							action(lam);
							lam.RemoveRight();
						}
					});
					graph.Nodes.Remove(lam);
				}
			}
			
			foreach (Node node in graph.Nodes.ToArray())
					action(node);
		}

		public static TermGraph CreateTree()
		{
			TermGraph tree = new TermGraph();

			//ESEMPIO PROF
			Node a4 = tree.AddVar();
			Node a3 = tree.AddApp(a4, a4);
			Node a2 = tree.AddApp(a3, a3);
			Node a1 = tree.AddLam(a2);
			a4.Binder = a1;
			Node b5 = tree.AddVar();
			Node b3 = tree.AddApp(b5, b5);
			Node b4 = tree.AddApp(b5, b5);
			Node b2 = tree.AddApp(b3, b4);
			Node b1 = tree.AddLam(b2);
			b5.Binder = b1;
			tree.AddUndirectEdge(a1, b1);

			//Node v1 = tree.AddVar();
			//Node v2 = tree.AddVar();
			//Node v3 = tree.AddVar();
			//Node v4 = tree.AddVar();
			//Node l2 = tree.AddLam(v2, v3);
			//Node l1 = tree.AddLam(v1, l2);
			//Node l3 = tree.AddLam(v4, v3);
			//Node a1 = tree.AddApp(l1, l2);
			//Node a2 = tree.AddApp(l3, l3);
			//tree.AddUndirectEdge(a1, a2);

			//Node v1 = tree.AddVar();
			//Node v2 = tree.AddVar();
			//Node v3 = tree.AddVar();
			//Node a1 = tree.AddApp(v1, v2);
			//Node a2 = tree.AddApp(v2, v2);
			//tree.AddUndirectEdge(a1, a2);

			//Node v1 = tree.AddVar();
			//Node v2 = tree.AddVar();
			//Node l1 = tree.AddLam(v1, v1);
			//Node a1 = tree.AddApp(v2, v2);
			//Node a2 = tree.AddApp(l1, l1);
			//Node a3 = tree.AddApp(a1, a1);
			//tree.AddUndirectEdge(a2, a3);

			//Node v1 = tree.AddVar();
			//Node v2 = tree.AddVar();
			//Node v3 = tree.AddVar();
			//Node l1 = tree.AddLam(v1, v1);
			//Node l2 = tree.AddLam(v2, v3);
			//tree.AddUndirectEdge(l1, l2);

			return tree;
		}
	}
}
